;;; checkgroups.el -- Do checkgroups via nntp.

;;; Copyright (C) 1997,1998 Katsumi Yamaoka
;;; Author:   Katsumi Yamaoka <yamaoka@jpl.org>
;;; Created:  97/07/10
;;; Revised:  98/11/17

;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2, or (at your option)
;;; any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with GNU Emacs; see the file COPYING.  If not, write to the
;;; Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;;; Boston, MA 02110-1301, USA.

(defconst checkgroups-version "0.3.4")

;;; Customizable variables.
(defvar checkgroups-news-server "flab"
  "*The name of the default host running nntp server. [checkgroups.el]")
  ;; Pay homage to Masanobu Umeda.

(defvar checkgroups-nntp-port "nntp"
  "*Nntp service name (\"nntp\" or 119). [checkgroups.el]")

(defvar checkgroups-comment-face 'bold
  "*Face to use for comments. [checkgroups.el]")

(defvar checkgroups-diff-face-! 'highlight
  "*Face to use for diff(!). [checkgroups.el]")

(defvar checkgroups-diff-face-+ 'highlight
  "*Face to use for diff(+). [checkgroups.el]")

(defvar checkgroups-diff-face-- 'highlight
  "*Face to use for diff(-). [checkgroups.el]")

(defvar checkgroups-diff-program "diff"
  "*DIFF executable. [checkgroups.el]")

(defvar checkgroups-temp-file
  (format "/tmp/.checkgroups-%d"
	  (if (fboundp 'emacs-pid) (emacs-pid) (user-real-uid)))
  "*Temp file name. [checkgroups.el]")

(defvar checkgroups-change-server-hook nil
  "*Hook run after the server changed. [checkgroups.el]")

(defvar checkgroups-load-hook nil
  "*Hook run while this package is loaded. [checkgroups.el]")

;;; Internal variables.
(defvar checkgroups-cmsg-active nil)
(defvar checkgroups-cmsg-newsgroups nil)
(defvar checkgroups-cmsg-newsgroups-normalized nil)
(defvar checkgroups-server-active nil)
(defvar checkgroups-server-newsgroups nil)
(defvar checkgroups-server-newsgroups-normalized nil)
(defconst checkgroups-e19-p (and (>= (string-to-int emacs-version) 19)
				 (fboundp 'put-text-property)))

;;; Internal functions.
(or checkgroups-e19-p
    (progn
      (fset 'defsubst (symbol-function 'defun))
      (fset 'buffer-substring-no-properties
	    (symbol-function 'buffer-substring))
      (fset 'put-text-property (function (lambda (&rest args) nil)))
      (setq checkgroups-comment-face nil
	    checkgroups-diff-face-! nil
	    checkgroups-diff-face-+ nil
	    checkgroups-diff-face-- nil)
      ))

(defun checkgroups-sort-function (a b)
  (string-lessp (car a) (car b)))

(defun checkgroups-insert (&rest args)
  (apply 'insert args)
  (sit-for 0))

(defmacro
  checkgroups-parse-cmsg-sub-1 (name-c name-a moderated name-n charter-n)
  (list 'setq
	'category (list 'append 'category (list 'list name-c))
	'checkgroups-cmsg-active (list 'append 'checkgroups-cmsg-active
				       (list 'list
					     (list 'cons name-a moderated)))
	'checkgroups-cmsg-newsgroups
	(list 'append 'checkgroups-cmsg-newsgroups
	      (list 'list (list 'cons name-n charter-n)))))

(defsubst checkgroups-parse-cmsg-sub-2 ()
  (not (not (re-search-forward "(Moderated)[\t ]*$" (match-end 0) t))))

(defmacro checkgroups-parse-cmsg-sub-3 (num)
  (list 'buffer-substring-no-properties 'beg (list 'match-end num)))

(defvar checkgroups-coding-system-for-decoding-string
  (cond ((fboundp 'decode-coding-string)
	 'undecided)
	((fboundp 'code-convert-string)
	 '*autoconv*)
	((fboundp 'convert-string-kanji-code)
	 2)))

(defmacro checkgroups-decode-coding-string (string coding-system)
  (cond ((fboundp 'decode-coding-string)
	 (list 'decode-coding-string string coding-system))
	((fboundp 'code-convert-string)
	 (list 'code-convert-string string coding-system '*internal*))
	((fboundp 'convert-string-kanji-code)
	 (list 'convert-string-kanji-code string convert-string-kanji-code 3))
	(t string)))

(defun checkgroups-parse-cmsg ()
  (let ((lines 0)
	lines2 beg end region name name2 category re buffer)
    (setq checkgroups-cmsg-active nil
	  checkgroups-cmsg-newsgroups nil
	  checkgroups-cmsg-newsgroups-normalized nil)
    (save-excursion
      (save-restriction
	(widen)
	(goto-char (point-min))
	(re-search-forward "\n\n+" nil t)
	(narrow-to-region (point) (point-max))
	(while (re-search-forward "^\\+\\([a-z][^\n\t .]*\\)" nil t)
	  (setq beg (match-beginning 1)
		name (checkgroups-parse-cmsg-sub-3 1))
	  (cond
	   ((looking-at "\\(\\.[^\n\t .]+\\)+")
	    (goto-char (match-end 0))
	    (setq name2 name
		  name (buffer-substring-no-properties beg (point)))
	    (cond
	     ((looking-at "[\t ]+.+$")
	      (checkgroups-parse-cmsg-sub-1
	       name2 name (checkgroups-parse-cmsg-sub-2) name
	       (checkgroups-parse-cmsg-sub-3 0)))
	     ((looking-at "[\t ]*$")
	      (checkgroups-parse-cmsg-sub-1 name2 name nil name name)))
	    )
	   ((looking-at "[\t ]+.+$")
	    (checkgroups-parse-cmsg-sub-1
	     name name (checkgroups-parse-cmsg-sub-2) name
	     (checkgroups-parse-cmsg-sub-3 0)))
	   ((looking-at "[\t ]+.+$")
	    (checkgroups-parse-cmsg-sub-1 name name nil name name))))
	(or
	 category
	 (progn
	   (goto-char (setq beg (point-min)))
	   (while (and (not (eobp)) (re-search-forward "^[\t ]*$" nil t))
	     (setq end (match-beginning 0)
		   lines2 (count-lines beg end))
	     (and (>= lines2 lines)
		  (setq lines lines2
			region (list beg end)))
	     (or (eobp)
		 (goto-char
		  (setq beg (or (and (re-search-forward "^[^\n\t ]" nil t)
				     (match-beginning 0))
				(point-max))))))
	   (or region (error "!! Parse error."))
	   (apply 'narrow-to-region region)
	   (goto-char (point-min))
	   (while (re-search-forward "^\\([a-z][^\n\t .]*\\)" nil t)
	     (setq beg (match-beginning 1)
		   name (checkgroups-parse-cmsg-sub-3 1))
	     (cond
	      ((looking-at "\\(\\.[^\n\t .]+\\)+")
	       (goto-char (match-end 0))
	       (setq name2 name
		     name (buffer-substring-no-properties beg (point)))
	       (cond
		((looking-at "[\t ]+.+$")
		 (checkgroups-parse-cmsg-sub-1
		  name2 name (checkgroups-parse-cmsg-sub-2)
		  name (checkgroups-parse-cmsg-sub-3 0)))
		((looking-at "[\t ]*$")
		 (checkgroups-parse-cmsg-sub-1 name2 name nil name name)))
	       )
	      ((looking-at "[\t ]+.+$")
	       (checkgroups-parse-cmsg-sub-1
		name name (checkgroups-parse-cmsg-sub-2)
		name (checkgroups-parse-cmsg-sub-3 0)))
	      ((looking-at "[\t ]*$")
	       (checkgroups-parse-cmsg-sub-1 name name nil name name))))))
	(or category (error "!! Parse error."))
	(setq category (sort category 'string-lessp))
	(or (string-equal (car category) (car (reverse category)))
	    (error "!! Parse error.")))
      (setq checkgroups-cmsg-active (sort
				     checkgroups-cmsg-active
				     'checkgroups-sort-function)
	    checkgroups-cmsg-newsgroups (sort
					 checkgroups-cmsg-newsgroups
					 'checkgroups-sort-function)
	    category (car category)
	    re (concat "^\\(" category "\\|" category
		       "\\.[^\n ]+\\)\\( \\| .+\\)?$")
	    buffer (get-buffer-create "*checkgroups*"))
      (set-buffer buffer)
      (setq buffer-read-only nil)
      (erase-buffer)
      (insert (mapconcat
	       (function (lambda (elt) (cdr elt)))
	       checkgroups-cmsg-newsgroups "\n") "\n")
      (goto-char (point-min))
      (while (re-search-forward "[\t ]+" nil t)
	(replace-match " "))
      (goto-char (point-min))
      (while (search-forward " \n" nil t)
	(replace-match "\n"))
      (goto-char (point-min))
      (while (re-search-forward re nil t)
	(setq checkgroups-cmsg-newsgroups-normalized
	      (append checkgroups-cmsg-newsgroups-normalized
		      (list (cons (buffer-substring-no-properties
				   (match-beginning 1) (match-end 1))
				  (buffer-substring-no-properties
				   (match-beginning 0) (match-end 0))))))))
    (kill-buffer buffer)
    category))

(defun checkgroups-get-active-and-newsgroups (category)
  (let ((buffer (get-buffer-create "*checkgroups*"))
	(category (regexp-quote category))
	(re1 "^\.\r?\n")
	(re2 "^2.+\.\r?\n")
	proc a-re n-re1 n-re2 a-beg a-end n-beg n-end)
    (setq a-re (concat "^\\(" category "\\|" category
		       "\\.[^\r\n\t ]+\\)"
		       "[\t ]+[0-9]+[\t ]+[0-9]+[\t ]+\\(y\\)?")
	  n-re1 (concat
		 "^\\(" category "\\|" category
		 "\\.[^\r\n\t ]+\\)\\([\t ]+\\|[\t ]+[^\r\n]+\\)?\\(\r?\n\\)")
	  n-re2 (concat "^\\(" category "\\|" category
			"\\.[^\n ]+\\)\\( \\| .+\\)?$"))
    (save-excursion
      (set-buffer buffer)
      (setq buffer-read-only nil)
      (erase-buffer)

      (message "Opening nntp connection to %s..." checkgroups-news-server)
      (setq proc (open-network-stream "*checkgroups*"
				      buffer
				      checkgroups-news-server
				      checkgroups-nntp-port))
      (cond ((fboundp 'set-process-kanji-code)		; NEmacs
	     (set-process-kanji-code proc 0))
	    ((fboundp 'set-process-coding-system)
	     (if (boundp 'MULE)
		 (set-process-coding-system		; Mule 2.x
		  proc '*noconv* '*noconv*)
	       (set-process-coding-system		; Mule 3.x, XEmacs
		proc 'binary 'binary)
	       )))
      (while (not (and (goto-char (point-min))
		       (re-search-forward re2 nil t)
		       (setq a-beg (match-end 0))))
	(accept-process-output))

      (message "Retrieving active file from %s via nntp..."
	       checkgroups-news-server)
      (process-send-string proc "LIST ACTIVE\r\n")
      (while (not (and (goto-char a-beg)
		       (re-search-forward re2 nil t)
		       (setq a-beg (match-end 0))))
	(accept-process-output))
      (while (not (and (goto-char a-beg)
		       (re-search-forward re1 nil t)
		       (setq a-end (match-beginning 0))))
	(accept-process-output))

      (message "Retrieving newsgroups file from %s via nntp..."
	       checkgroups-news-server)
      (goto-char (setq n-beg (point-max)))
      (process-send-string proc "LIST NEWSGROUPS\r\n")
      (while (not (and (goto-char n-beg)
		       (re-search-forward re2 nil t)
		       (setq n-beg (match-end 0))))
	(accept-process-output))
      (while (not (and (goto-char n-beg)
		       (re-search-forward re1 nil t)
		       (setq n-end (match-beginning 0))))
	(accept-process-output))

      (process-send-string proc "QUIT\r\n")

      (message "Parsing active file...")
      (goto-char a-beg)
      (setq checkgroups-server-active nil)
      (while (re-search-forward a-re a-end t)
	(setq checkgroups-server-active
	      (append checkgroups-server-active
		      (list (cons (buffer-substring-no-properties
				   (match-beginning 1) (match-end 1))
				  (not (match-beginning 2)))))))
      (setq checkgroups-server-active (sort checkgroups-server-active
					    'checkgroups-sort-function))

      (message "Parsing newsgroups file...")
      (goto-char n-beg)
      (setq checkgroups-server-newsgroups nil)
      (while (re-search-forward n-re1 n-end t)
	(setq checkgroups-server-newsgroups
	      (append checkgroups-server-newsgroups
		      (list
		       (cons
			(buffer-substring-no-properties
			 (match-beginning 1) (match-end 1))
			(checkgroups-decode-coding-string
			 (buffer-substring-no-properties
			  (match-beginning 0)
			  (match-beginning 3))
			 checkgroups-coding-system-for-decoding-string))))))
      (setq checkgroups-server-newsgroups (sort checkgroups-server-newsgroups
						'checkgroups-sort-function))

      (message "Parsing normalized newsgroups...")
      (erase-buffer)
      (insert (mapconcat
	       (function (lambda (elt) (cdr elt)))
	       checkgroups-server-newsgroups "\n") "\n")
      (goto-char (point-min))
      (while (re-search-forward "[\t ]+" nil t)
	(replace-match " "))
      (goto-char (point-min))
      (while (search-forward " \n" nil t)
	(replace-match "\n"))
      (goto-char (point-min))
      (setq checkgroups-server-newsgroups-normalized nil)
      (while (re-search-forward n-re2 nil t)
	(setq checkgroups-server-newsgroups-normalized
	      (append checkgroups-server-newsgroups-normalized
		      (list (cons (buffer-substring-no-properties
				   (match-beginning 1) (match-end 1))
				  (buffer-substring-no-properties
				   (match-beginning 0) (match-end 0)))))))
      (message "Parsing normalized newsgroups...done")
      )
    (kill-buffer buffer)
    ))

;;; Commands.
(defun checkgroups-change-server (server)
  (interactive (list (read-string "Server: " checkgroups-news-server)))
  (setq checkgroups-news-server server)
  (run-hooks 'checkgroups-change-server-hook))

(defun checkgroups-fine-diff-sub-1 (str suffix)
  (erase-buffer)
  (insert str)
  (goto-char (point-min))
  (re-search-forward "[\t ]+\\|$" nil t)
  (while (not (eobp))
    (insert "\n")
    (forward-char 1))
  (insert "\n;;; END-" suffix "\n")
  (goto-char (point-min))
  (insert ";;; BEG-" suffix "\n")
  (count-lines (point-min) (point-max)))

(defun checkgroups-fine-diff-sub-2 (beg end prefix)
  (let ((faces '((?! . checkgroups-diff-face-!)
		 (?+ . checkgroups-diff-face-+)
		 (?- . checkgroups-diff-face--))))
    (save-restriction
      (narrow-to-region beg end)
      (goto-char (point-min))
      (while (re-search-forward "^[!+-] \\(.\\)$" nil t)
	(put-text-property
	 (match-beginning 1) (match-end 1) 'face
	 (symbol-value (cdr (assq (char-after (match-beginning 0)) faces)))))
      (goto-char (point-min))
      (while (not (eobp))
	(delete-char 2)
	(forward-line 1))
      (goto-char (point-min))
      (insert prefix)
      (while (search-forward "\n" nil t)
	(replace-match ""))
      (goto-char (point-max))
      (insert "\n"))))

(defun checkgroups-fine-diff-strings (str-A str-B buffer)
  (let (lines)
    (save-excursion
      (set-buffer buffer)
      (setq buffer-read-only nil)
      (setq lines (checkgroups-fine-diff-sub-1 str-A "A"))
      (unwind-protect
	  (let (
		; NEmacs
		(default-kanji-process-code 3)
		(kanji-fileio-code 0)
		program-kanji-code-alist
		; Mule-{1,2}.x
		(default-process-coding-system '(*euc-japan* . *euc-japan*))
		(file-coding-system '*euc-japan*)
		program-coding-system-alist
		; XEmacs-20.x
		(buffer-file-coding-system 'euc-japan)
		(coding-system-for-read 'euc-japan)
		(coding-system-for-write 'euc-japan)
		)
	    (write-region (point-min) (point-max) checkgroups-temp-file
			  nil 'quiet)
	    (setq lines (max lines (checkgroups-fine-diff-sub-1 str-B "B")))
	    (call-process-region
	     (point-min) (point-max) checkgroups-diff-program
	     t t nil (format "-c%d" lines) "-w" checkgroups-temp-file "-"))
	(delete-file checkgroups-temp-file))
      (goto-char (point-min))
      (re-search-forward "^. ;;; BEG-.\n")
      (delete-region (point-min) (point))
      (delete-region (progn (re-search-forward "^. ;;; END-.\n")
			    (match-beginning 0))
		     (progn (re-search-forward "^. ;;; BEG-.\n")
			    (point)))
      (checkgroups-fine-diff-sub-2 (point-min) (point) "-")
      (checkgroups-fine-diff-sub-2
       (point)
       (progn (re-search-forward "^. ;;; END-.\n")
	      (delete-region (match-beginning 0) (point-max))
	      (point-max))
       "+")
      buffer)))

(defun checkgroups (arg)
  "Do checkgroups. [checkgroups.el]"
  (interactive "P")
  (let ((checkgroups-news-server checkgroups-news-server)
	(obuf (current-buffer))
	(gc-cons-threshold (if (< 8388608 0) 4194303 8388607))
	(imu 0)
	(imm 0)
	(missings 0)
	(excesses 0)
	(diff 0)
	category start1 start2 cactive ca mod elt sactive sa cng sng buffer
	sngs)
    (and arg (call-interactively 'checkgroups-change-server))
    (message "Parsing reference...")
    (save-excursion
      (and (boundp 'gnus-summary-buffer)
	   (eq (current-buffer) (get-buffer gnus-summary-buffer))
	   (progn
	     (gnus-summary-show-article)
	     (message "Parsing reference...")
	     (set-buffer (get-buffer gnus-article-buffer))))
      (setq category (checkgroups-parse-cmsg)))
    (message "Parsing reference...done")
    (checkgroups-get-active-and-newsgroups category)
    (switch-to-buffer obuf)
    (switch-to-buffer (get-buffer-create "*checkgroups*"))
    (delete-other-windows)
    (setq buffer-read-only nil)
    (erase-buffer)

    (insert "# The following newsgroups were missing and should be added.\n")
    (and checkgroups-comment-face

	 (put-text-property (point-min) (1- (point))
			    'face checkgroups-comment-face))
    (setq start1 (point))
    (insert "\n# The following groups are incorrectly marked as moderated.\n")
    (and checkgroups-comment-face
	 (put-text-property (1+ start1) (1- (point))
			    'face checkgroups-comment-face))
    (setq start2 (point))
    (insert "\n"
	    "# The following groups are incorrectly marked as unmoderated.\n")
    (and checkgroups-comment-face
	 (put-text-property (1+ start2) (1- (point))
			    'face checkgroups-comment-face))
    (message "Comparing active file...")
    (setq cactive checkgroups-cmsg-active)
    (while cactive
      (setq ca (car (car cactive))
	    mod (cdr (car cactive))
	    cactive (cdr cactive)
	    elt (assoc ca checkgroups-server-active))
      (if elt
	  (cond ((and mod (not (cdr elt)))
		 (setq imu (1+ imu))
		 (goto-char (point-max))
		 (checkgroups-insert
		  (cdr (assoc ca checkgroups-cmsg-newsgroups)) "\n"))
		((and (not mod) (cdr elt))
		 (setq imm (1+ imm))
		 (goto-char start2)
		 (checkgroups-insert
		  (cdr (assoc ca checkgroups-cmsg-newsgroups)) "\n")
		 (setq start2 (point))))
	(setq missings (1+ missings))
	(goto-char start1)
	(checkgroups-insert (cdr (assoc ca checkgroups-cmsg-newsgroups)) "\n")
	(setq start2 (+ (- (point) start1) start2)
	      start1 (point))))
    (and (zerop imu)
	 (progn
	   (delete-region start2 (point-max))
	   (setq start2 (point-max))))
    (and (zerop imm) (delete-region start1 start2))
    (and (zerop missings)
	 (delete-region (point-min) (min (1+ start1) (point-max))))

    (set-buffer-modified-p nil)
    (goto-char (point-min))
    (sit-for 0)

    (message "Comparing newsgroups file...")
    (goto-char (point-max))
    (setq start1 (point))
    (or (bobp) (insert "\n"))
    (setq start2 (point))
    (insert "# The following newsgroups are non-standard.\n")
    (and checkgroups-comment-face
	 (put-text-property start2 (1- (point))
			    'face checkgroups-comment-face))
    (setq sactive checkgroups-server-active)
    (while sactive
      (setq sa (car (car sactive))
	    sactive (cdr sactive))
      (or (assoc sa checkgroups-cmsg-active)
	  (progn
	    (setq excesses (1+ excesses))
	    (checkgroups-insert
	     (or (cdr (assoc sa checkgroups-server-newsgroups)) sa) "\n"))))
    (and (zerop excesses) (delete-region start1 (point-max)))

    (goto-char (point-max))
    (setq start1 (point))
    (or (bobp) (insert "\n"))
    (setq start2 (point))
    (insert "# The following charters are missing"
	    " or different from the reference.\n"
	    "# -{newsgroups file}\n"
	    "# +{reference}\n")
    (and checkgroups-comment-face
	 (put-text-property start2 (1- (point))
			    'face checkgroups-comment-face))

    (setq cactive checkgroups-cmsg-active)
    (if (and checkgroups-e19-p
	     (or checkgroups-diff-face-!
		 checkgroups-diff-face-+
		 checkgroups-diff-face--))
	(progn
	  (message "Computing fine diff...")
	  (setq buffer (get-buffer-create " *checkgroups-temp*"))
	  (while cactive
	    (setq ca (car (car cactive))
		  cactive (cdr cactive)
		  cng (cdr (assoc ca checkgroups-cmsg-newsgroups-normalized))
		  sng (or
		       (cdr
			(assoc ca checkgroups-server-newsgroups-normalized))
		       ""))
	    (or (string-equal cng sng)
		(progn
		  (setq diff (1+ diff))
		  (insert "\n")
		  (insert-buffer-substring
		   (checkgroups-fine-diff-strings
		    (or (cdr (assoc ca checkgroups-server-newsgroups)) "")
		    (cdr (assoc ca checkgroups-cmsg-newsgroups))
		    buffer))
		  (goto-char (point-max))
		  (sit-for 0)
		  )))
	  (kill-buffer buffer))
      (while cactive
	(setq ca (car (car cactive))
	      cactive (cdr cactive)
	      cng (cdr (assoc ca checkgroups-cmsg-newsgroups-normalized))
	      sng (or (cdr (assoc ca checkgroups-server-newsgroups-normalized))
		      ""))
	(or (string-equal cng sng)
	    (progn
	      (setq diff (1+ diff))
	      (checkgroups-insert
	       "\n-"
	       (or (cdr (assoc ca checkgroups-server-newsgroups)) "")
	       "\n+" (cdr (assoc ca checkgroups-cmsg-newsgroups)) "\n"
	       ))))
      )
    (and (zerop diff) (delete-region start1 (point-max)))

    (setq start1 (point))
    (insert "\n# The following lines in the newsgroups file"
	    " are non-standard.\n")
    (and checkgroups-comment-face
	 (put-text-property (1+ start1) (1- (point))
			    'face checkgroups-comment-face))
    (message "Checking newsgroups file...")
    (setq sngs checkgroups-server-newsgroups
	  excesses 0)
    (while sngs
      (setq sng (car sngs)
	    sngs (cdr sngs))
      (or (assoc (car sng) checkgroups-cmsg-active)
	  (progn
	    (setq excesses (1+ excesses))
	    (checkgroups-insert (cdr sng) "\n"))))
    (and (zerop excesses) (delete-region start1 (point-max)))
    (set-buffer-modified-p nil)
    (setq buffer-read-only t)
    (goto-char (point-min))
    (message "Done")
    ))

(provide 'checkgroups)
(run-hooks 'checkgroups-load-hook)

;;; checkgroups.el ends here
